#include "decoder.h"

UNIFEX_TERM create_result_ok(UnifexEnv *env, UnifexState *state) {
  return ({
    const ERL_NIF_TERM terms[] = {enif_make_atom(env, "ok"),
                                  unifex_make_resource(env, state)};
    enif_make_tuple_from_array(env, terms, 2);
  });
}

UNIFEX_TERM create_result_error_unknown(UnifexEnv *env) {
  return ({
    const ERL_NIF_TERM terms[] = {enif_make_atom(env, "error"),
                                  enif_make_atom(env, "unknown")};
    enif_make_tuple_from_array(env, terms, 2);
  });
}

UNIFEX_TERM create_result_error_no_memory(UnifexEnv *env) {
  return ({
    const ERL_NIF_TERM terms[] = {enif_make_atom(env, "error"),
                                  enif_make_atom(env, "no_memory")};
    enif_make_tuple_from_array(env, terms, 2);
  });
}

UNIFEX_TERM get_metadata_result(UnifexEnv *env, long frame_size,
                                long sample_rate, int channels) {
  return ({
    const ERL_NIF_TERM terms[] = {enif_make_atom(env, "ok"), ({
                                    const ERL_NIF_TERM terms[] = {
                                        enif_make_long(env, frame_size),
                                        enif_make_long(env, sample_rate),
                                        enif_make_int(env, channels)};
                                    enif_make_tuple_from_array(env, terms, 3);
                                  })};
    enif_make_tuple_from_array(env, terms, 2);
  });
}

UNIFEX_TERM fill_result_ok(UnifexEnv *env) { return enif_make_atom(env, "ok"); }

UNIFEX_TERM fill_result_error_invalid_data(UnifexEnv *env) {
  return ({
    const ERL_NIF_TERM terms[] = {enif_make_atom(env, "error"),
                                  enif_make_atom(env, "invalid_data")};
    enif_make_tuple_from_array(env, terms, 2);
  });
}

UNIFEX_TERM decode_frame_result_ok(UnifexEnv *env, UnifexPayload *payload) {
  return ({
    const ERL_NIF_TERM terms[] = {enif_make_atom(env, "ok"),
                                  unifex_payload_to_term(env, payload)};
    enif_make_tuple_from_array(env, terms, 2);
  });
}

UNIFEX_TERM decode_frame_result_error_not_enough_bits(UnifexEnv *env) {
  return ({
    const ERL_NIF_TERM terms[] = {enif_make_atom(env, "error"),
                                  enif_make_atom(env, "not_enough_bits")};
    enif_make_tuple_from_array(env, terms, 2);
  });
}

UNIFEX_TERM decode_frame_result_error_unknown(UnifexEnv *env) {
  return ({
    const ERL_NIF_TERM terms[] = {enif_make_atom(env, "error"),
                                  enif_make_atom(env, "unknown")};
    enif_make_tuple_from_array(env, terms, 2);
  });
}

ErlNifResourceType *STATE_RESOURCE_TYPE;

UnifexState *unifex_alloc_state(UnifexEnv *env) {
  UNIFEX_UNUSED(env);
  return (UnifexState *)enif_alloc_resource(STATE_RESOURCE_TYPE,
                                            sizeof(UnifexState));
}

void unifex_release_state(UnifexEnv *env, UnifexState *state) {
  UNIFEX_UNUSED(env);
  enif_release_resource(state);
}

void unifex_keep_state(UnifexEnv *env, UnifexState *state) {
  UNIFEX_UNUSED(env);
  enif_keep_resource(state);
}

static void destroy_state(ErlNifEnv *env, void *value) {
  UnifexState *state = (UnifexState *)value;
  UnifexEnv *unifex_env = env;
  handle_destroy_state(unifex_env, state);
}

static int unifex_load_nif(ErlNifEnv *env, void **priv_data,
                           ERL_NIF_TERM load_info) {
  UNIFEX_UNUSED(load_info);
  UNIFEX_UNUSED(priv_data);

  ErlNifResourceFlags flags =
      (ErlNifResourceFlags)(ERL_NIF_RT_CREATE | ERL_NIF_RT_TAKEOVER);

  STATE_RESOURCE_TYPE =
      enif_open_resource_type(env, NULL, "UnifexState",
                              (ErlNifResourceDtor *)destroy_state, flags, NULL);

  UNIFEX_PAYLOAD_GUARD_RESOURCE_TYPE = enif_open_resource_type(
      env, NULL, "UnifexPayloadGuard",
      (ErlNifResourceDtor *)unifex_payload_guard_destructor, flags, NULL);

  return 0;
}

static ERL_NIF_TERM export_create(ErlNifEnv *env, int argc,
                                  const ERL_NIF_TERM argv[]) {
  UNIFEX_MAYBE_UNUSED(argc);
  UNIFEX_MAYBE_UNUSED(argv);
  ERL_NIF_TERM result;
  UnifexEnv *unifex_env = env;

  result = create(unifex_env);
  goto exit_export_create;
exit_export_create:

  return result;
}

static ERL_NIF_TERM export_get_metadata(ErlNifEnv *env, int argc,
                                        const ERL_NIF_TERM argv[]) {
  UNIFEX_MAYBE_UNUSED(argc);
  UNIFEX_MAYBE_UNUSED(argv);
  ERL_NIF_TERM result;
  UnifexEnv *unifex_env = env;
  UnifexState *state;

  if (!enif_get_resource(env, argv[0], STATE_RESOURCE_TYPE, (void **)&state)) {
    result = unifex_raise_args_error(env, "state", ":state");
    goto exit_export_get_metadata;
  }

  result = get_metadata(unifex_env, state);
  goto exit_export_get_metadata;
exit_export_get_metadata:

  return result;
}

static ERL_NIF_TERM export_fill(ErlNifEnv *env, int argc,
                                const ERL_NIF_TERM argv[]) {
  UNIFEX_MAYBE_UNUSED(argc);
  UNIFEX_MAYBE_UNUSED(argv);
  ERL_NIF_TERM result;
  UnifexEnv *unifex_env = env;
  UnifexPayload *payload;
  UnifexState *state;

  payload = (UnifexPayload *)unifex_alloc(sizeof(UnifexPayload));

  if (!unifex_payload_from_term(env, argv[0], payload)) {
    result = unifex_raise_args_error(env, "payload", ":payload");
    goto exit_export_fill;
  }

  if (!enif_get_resource(env, argv[1], STATE_RESOURCE_TYPE, (void **)&state)) {
    result = unifex_raise_args_error(env, "state", ":state");
    goto exit_export_fill;
  }

  result = fill(unifex_env, payload, state);
  goto exit_export_fill;
exit_export_fill:
  unifex_payload_release_ptr(&payload);
  return result;
}

static ERL_NIF_TERM export_decode_frame(ErlNifEnv *env, int argc,
                                        const ERL_NIF_TERM argv[]) {
  UNIFEX_MAYBE_UNUSED(argc);
  UNIFEX_MAYBE_UNUSED(argv);
  ERL_NIF_TERM result;
  UnifexEnv *unifex_env = env;
  UnifexPayload *payload;
  UnifexState *state;

  payload = (UnifexPayload *)unifex_alloc(sizeof(UnifexPayload));

  if (!unifex_payload_from_term(env, argv[0], payload)) {
    result = unifex_raise_args_error(env, "payload", ":payload");
    goto exit_export_decode_frame;
  }

  if (!enif_get_resource(env, argv[1], STATE_RESOURCE_TYPE, (void **)&state)) {
    result = unifex_raise_args_error(env, "state", ":state");
    goto exit_export_decode_frame;
  }

  result = decode_frame(unifex_env, payload, state);
  goto exit_export_decode_frame;
exit_export_decode_frame:
  unifex_payload_release_ptr(&payload);
  return result;
}

static ErlNifFunc nif_funcs[] = {
    {"unifex_create", 0, export_create, 0},
    {"unifex_get_metadata", 1, export_get_metadata, 0},
    {"unifex_fill", 2, export_fill, 0},
    {"unifex_decode_frame", 2, export_decode_frame, 0}};

ERL_NIF_INIT(Elixir.Membrane.AAC.FDK.Decoder.Native.Nif, nif_funcs,
             unifex_load_nif, NULL, NULL, NULL)
