#include "encoder.h"

UNIFEX_TERM create_result_ok(UnifexEnv *env, UnifexState *state) {
  return ({
    const ERL_NIF_TERM terms[] = {enif_make_atom(env, "ok"),
                                  unifex_make_resource(env, state)};
    enif_make_tuple_from_array(env, terms, 2);
  });
}

UNIFEX_TERM create_result_error(UnifexEnv *env, const char *reason) {
  return ({
    const ERL_NIF_TERM terms[] = {enif_make_atom(env, "error"),
                                  enif_make_atom(env, reason)};
    enif_make_tuple_from_array(env, terms, 2);
  });
}

UNIFEX_TERM encode_frame_result_ok(UnifexEnv *env, UnifexPayload *payload) {
  return ({
    const ERL_NIF_TERM terms[] = {enif_make_atom(env, "ok"),
                                  unifex_payload_to_term(env, payload)};
    enif_make_tuple_from_array(env, terms, 2);
  });
}

UNIFEX_TERM encode_frame_result_error(UnifexEnv *env, const char *reason) {
  return ({
    const ERL_NIF_TERM terms[] = {enif_make_atom(env, "error"),
                                  enif_make_atom(env, reason)};
    enif_make_tuple_from_array(env, terms, 2);
  });
}

UNIFEX_TERM encode_frame_result_error_no_data(UnifexEnv *env) {
  return ({
    const ERL_NIF_TERM terms[] = {enif_make_atom(env, "error"),
                                  enif_make_atom(env, "no_data")};
    enif_make_tuple_from_array(env, terms, 2);
  });
}

ErlNifResourceType *STATE_RESOURCE_TYPE;

UnifexState *unifex_alloc_state(UnifexEnv *env) {
  UNIFEX_UNUSED(env);
  return (UnifexState *)enif_alloc_resource(STATE_RESOURCE_TYPE,
                                            sizeof(UnifexState));
}

void unifex_release_state(UnifexEnv *env, UnifexState *state) {
  UNIFEX_UNUSED(env);
  enif_release_resource(state);
}

void unifex_keep_state(UnifexEnv *env, UnifexState *state) {
  UNIFEX_UNUSED(env);
  enif_keep_resource(state);
}

static void destroy_state(ErlNifEnv *env, void *value) {
  UnifexState *state = (UnifexState *)value;
  UnifexEnv *unifex_env = env;
  handle_destroy_state(unifex_env, state);
}

static int unifex_load_nif(ErlNifEnv *env, void **priv_data,
                           ERL_NIF_TERM load_info) {
  UNIFEX_UNUSED(load_info);
  UNIFEX_UNUSED(priv_data);

  ErlNifResourceFlags flags =
      (ErlNifResourceFlags)(ERL_NIF_RT_CREATE | ERL_NIF_RT_TAKEOVER);

  STATE_RESOURCE_TYPE =
      enif_open_resource_type(env, NULL, "UnifexState",
                              (ErlNifResourceDtor *)destroy_state, flags, NULL);

  UNIFEX_PAYLOAD_GUARD_RESOURCE_TYPE = enif_open_resource_type(
      env, NULL, "UnifexPayloadGuard",
      (ErlNifResourceDtor *)unifex_payload_guard_destructor, flags, NULL);

  return 0;
}

static ERL_NIF_TERM export_create(ErlNifEnv *env, int argc,
                                  const ERL_NIF_TERM argv[]) {
  UNIFEX_MAYBE_UNUSED(argc);
  UNIFEX_MAYBE_UNUSED(argv);
  ERL_NIF_TERM result;
  UnifexEnv *unifex_env = env;
  int channels;
  int sample_rate;
  int aot;
  int bitrate_mode;
  int bitrate;

  if (!enif_get_int(env, argv[0], &channels)) {
    result = unifex_raise_args_error(env, "channels", ":int");
    goto exit_export_create;
  }

  if (!enif_get_int(env, argv[1], &sample_rate)) {
    result = unifex_raise_args_error(env, "sample_rate", ":int");
    goto exit_export_create;
  }

  if (!enif_get_int(env, argv[2], &aot)) {
    result = unifex_raise_args_error(env, "aot", ":int");
    goto exit_export_create;
  }

  if (!enif_get_int(env, argv[3], &bitrate_mode)) {
    result = unifex_raise_args_error(env, "bitrate_mode", ":int");
    goto exit_export_create;
  }

  if (!enif_get_int(env, argv[4], &bitrate)) {
    result = unifex_raise_args_error(env, "bitrate", ":int");
    goto exit_export_create;
  }

  result =
      create(unifex_env, channels, sample_rate, aot, bitrate_mode, bitrate);
  goto exit_export_create;
exit_export_create:

  return result;
}

static ERL_NIF_TERM export_encode_frame(ErlNifEnv *env, int argc,
                                        const ERL_NIF_TERM argv[]) {
  UNIFEX_MAYBE_UNUSED(argc);
  UNIFEX_MAYBE_UNUSED(argv);
  ERL_NIF_TERM result;
  UnifexEnv *unifex_env = env;
  UnifexPayload *payload;
  UnifexState *state;

  payload = (UnifexPayload *)unifex_alloc(sizeof(UnifexPayload));

  if (!unifex_payload_from_term(env, argv[0], payload)) {
    result = unifex_raise_args_error(env, "payload", ":payload");
    goto exit_export_encode_frame;
  }

  if (!enif_get_resource(env, argv[1], STATE_RESOURCE_TYPE, (void **)&state)) {
    result = unifex_raise_args_error(env, "state", ":state");
    goto exit_export_encode_frame;
  }

  result = encode_frame(unifex_env, payload, state);
  goto exit_export_encode_frame;
exit_export_encode_frame:
  unifex_payload_release_ptr(&payload);
  return result;
}

static ErlNifFunc nif_funcs[] = {
    {"unifex_create", 5, export_create, 0},
    {"unifex_encode_frame", 2, export_encode_frame, 0}};

ERL_NIF_INIT(Elixir.Membrane.AAC.FDK.Encoder.Native.Nif, nif_funcs,
             unifex_load_nif, NULL, NULL, NULL)
