defmodule Membrane.Element.FDK.AAC.MixProject do
  use Mix.Project

  @version "0.2.2"
  @github_url "https://github.com/membraneframework/membrane-element-fdk-aac"

  def project do
    [
      app: :membrane_element_fdk_aac,
      # add unifex and bundlex to compilers
      compilers: [:unifex, :bundlex] ++ Mix.compilers(),
      version: @version,
      elixir: "~> 1.7",
      elixirc_paths: elixirc_paths(Mix.env()),
      start_permanent: Mix.env() == :prod,
      description: "Membrane Multimedia Framework (FDK-AAC Element)",
      package: package(),
      name: "Membrane Element: FDK-AAC",
      source_url: @github_url,
      docs: docs(),
      homepage_url: "https://membraneframework.org",
      deps: deps()
    ]
  end

  def application do
    [
      extra_applications: []
    ]
  end

  defp elixirc_paths(:test), do: ["lib", "test/support"]
  defp elixirc_paths(_env), do: ["lib"]

  defp docs do
    [
      main: "readme",
      extras: ["README.md"],
      source_ref: "v#{@version}"
    ]
  end

  defp package do
    [
      maintainers: ["Membrane Team"],
      licenses: ["Apache 2.0"],
      links: %{
        "GitHub" => @github_url,
        "Membrane Framework Homepage" => "https://membraneframework.org"
      },
      files: ["lib", "mix.exs", "README*", "LICENSE*", ".formatter.exs", "bundlex.exs", "c_src"]
    ]
  end

  defp deps do
    [
      {:membrane_core, "~> 0.5.0"},
      {:membrane_common_c, "~> 0.3.0"},
      {:bunch, "~> 1.0"},
      {:ex_doc, "~> 0.19", only: :dev, runtime: false},
      {:dialyxir, "~> 1.0.0-rc.4", only: [:dev], runtime: false},
      {:unifex, "~> 0.2.3"},
      {:membrane_caps_audio_raw, "~> 0.2.0"},
      {:membrane_element_file, "~> 0.3.0", only: [:dev, :test]}
    ]
  end
end
